//---------------------------------------------------------------------------
#include <vcl.h>
#pragma hdrstop

#include "dtUnit1.h"
//---------------------------------------------------------------------------
#pragma package(smart_init)
#pragma resource "*.dfm"
TForm1 *Form1;
const String IniSection = "dt";
//---------------------------------------------------------------------------
__fastcall TForm1::TForm1(TComponent* Owner)
    : TForm(Owner)
{
    Form2 = 0;
}
//---------------------------------------------------------------------------
String TForm1::GetIniString(const String Ident, const String Default){
    TIniFile *ini=new TIniFile(ChangeFileExt(Application->ExeName, ".set"));
    String retVal=ini->ReadString(IniSection, Ident, Default);
    delete ini;
    return retVal;
}
//---------------------------------------------------------------------------
void TForm1::PutIniString(const String Ident, const String Value){
    TIniFile *ini=new TIniFile(ChangeFileExt(Application->ExeName, ".set"));
    ini->WriteString(IniSection, Ident, Value);
    delete ini;
}
//---------------------------------------------------------------------------
void TForm1::LogOpen(){
    String fileName = ChangeFileExt(Application->ExeName, ".log");
    String backup = ChangeFileExt(Application->ExeName, ".old");
    if(FileExists(fileName)){
        if(FileExists(backup))
            DeleteFile(backup);
        RenameFile(fileName, backup);
    }
    try{
        iLogfile = FileCreate(fileName);
        LogWrite("Program start");
    }
    catch(...){
    }
}
//---------------------------------------------------------------------------
void TForm1::LogWrite(const String Value){
    String buf = DateTimeToStr(Now()) + ": " + Value + "\r\n";
    try{
        FileWrite(iLogfile, buf.c_str(), buf.Length());
    }
    catch(...){
    }
}
//---------------------------------------------------------------------------
void TForm1::LogClose(){
    try{
        LogWrite("Program stop");
        FileClose(iLogfile);
    }
    catch(...){
    }
}
//---------------------------------------------------------------------------
void __fastcall TForm1::FormCreate(TObject *Sender)
{
    // open logfile
    LogOpen();
    // install ourselves if needed
    Install(GetIniString("autostart", "1") == "1");
    // register custom message
    iconMsg = RegisterWindowMessage("{3482CDC0-E9AC-11D2-ACDE-E38E0A0C951D}");
    // put our icon in the tray area
    NOTIFYICONDATA icondata;
    ZeroMemory(&icondata, sizeof(NOTIFYICONDATA));
    icondata.cbSize = sizeof(NOTIFYICONDATA);
    icondata.hWnd = Handle;
    lstrcpy(icondata.szTip, "Daytime client");
    icondata.uCallbackMessage = iconMsg;
    icondata.uFlags = NIF_MESSAGE | NIF_ICON | NIF_TIP;
    icondata.hIcon = LoadImage(HInstance, "T_ICON", IMAGE_ICON, 16, 16, 0);
    Shell_NotifyIcon(NIM_ADD, &icondata);
    ComputeBumpUp();
    localIP = GetIniString("local IP", "127.0.0.1");
    iDelay = StrToIntDef(GetIniString("initial delay", "10"), 10);
    Timer1->Interval = StrToIntDef(GetIniString("timer interval",
            "5"), 5) * 1000;
    // start timer
    Timer1->Enabled = true;
}
//---------------------------------------------------------------------------
void __fastcall TForm1::FormDestroy(TObject *Sender)
{
    // remove our icon from tray area
    NOTIFYICONDATA icondata;
    ZeroMemory(&icondata, sizeof(NOTIFYICONDATA));
    icondata.cbSize = sizeof(NOTIFYICONDATA);
    icondata.hWnd = Handle;
    Shell_NotifyIcon(NIM_DELETE, &icondata);
    // close logfile
    LogClose();
}
//---------------------------------------------------------------------------
void __fastcall TForm1::WndProc(Messages::TMessage &msg){
    if(msg.Msg == iconMsg){
        switch(msg.LParam){
            case WM_LBUTTONDOWN:
                DoDialog();
                return;
            case WM_RBUTTONDOWN:
                if(!Timer1->Enabled) // busy with a dialog
                    return;
                Timer1->Enabled = false; // stall processing
                if(Application->MessageBox("Sure?", "QUIT",
                        MB_YESNO|MB_TOPMOST) == IDYES)
                    Application->Terminate();
                Timer1->Enabled = true; // continue processing
                return;
        }
    }
    TForm::WndProc(msg);
}
//---------------------------------------------------------------------------
void TForm1::Install(const bool bAction){
    // Install in autorun; bAction is install/remove
    const String ourID = "{73B861E0-EB45-11D2-ACE0-9933351E4818}";
    String oldValue;
    TRegistry *reg = new TRegistry();
    try{
        reg->RootKey = HKEY_LOCAL_MACHINE;
        reg->OpenKey("SOFTWARE\\Microsoft\\Windows\\CurrentVersion\\Run", true);
        if(reg->ValueExists(ourID))
            oldValue = reg->ReadString(ourID);
        if(bAction){
            if(oldValue != Application->ExeName)
                reg->WriteString(ourID, Application->ExeName);
        }
        else{
            if(oldValue.Length())
                reg->DeleteValue(ourID);
        }
        reg->CloseKey();
    }
    catch(...){
    }
    delete reg;
}
//---------------------------------------------------------------------------
void __fastcall TForm1::NMTime1Connect(TObject *Sender)
{
    LogWrite("Connect");
}
//---------------------------------------------------------------------------
void __fastcall TForm1::NMTime1ConnectionFailed(TObject *Sender)
{
    LogWrite("Connection failed");
}
//---------------------------------------------------------------------------
void __fastcall TForm1::NMTime1Disconnect(TObject *Sender)
{
    LogWrite("Disconnect");
}
//---------------------------------------------------------------------------
void __fastcall TForm1::NMTime1HostResolved(TComponent *Sender)
{
    LogWrite("Host resolved");
}
//---------------------------------------------------------------------------
void __fastcall TForm1::NMTime1InvalidHost(bool &Handled)
{
    LogWrite("Invalid host");
    Handled = false; // raise exception
}
//---------------------------------------------------------------------------
void __fastcall TForm1::NMTime1Status(TComponent *Sender,
      AnsiString Status)
{
    LogWrite(Status);
}
//---------------------------------------------------------------------------
void __fastcall TForm1::Timer1Timer(TObject *Sender)
{
    iDelay -= (Timer1->Interval / 1000);
    if(iDelay < 0)
        iDelay = 0;
    if(iDelay)
        return;
    if(localIP != NMTime1->LocalIP){
        Timer1->Enabled = false; // prevent double processing
        LogWrite("Connection detected: " + NMTime1->LocalIP);
        FILETIME ft;
        SYSTEMTIME st;
        unsigned __int64 qwTimeInt, qwFinal;
        ULARGE_INTEGER uli;
        // retrieve RFC 868 seconds since 01/01/1900
        // (this will rollover in 2036)
        try{
            NMTime1->TimeOut = StrToIntDef(GetIniString("NM timeout",
                    "30"), 30) * 1000;
            NMTime1->Port = 37; // std port for RFC 868
            TStringList *hosts = new TStringList;
            hosts->CommaText = GetIniString("host IPs", "129.6.16.35");
            int hostIndex = StrToIntDef(GetIniString("current host", "0"), 0);
            if(hostIndex >= hosts->Count)
                hostIndex = 0; // avoid out of bounds
            NMTime1->Host = hosts->Strings[hostIndex];
            delete hosts;
            LogWrite("About to talk to daytime host " + NMTime1->Host);
            NMTime1->ReportLevel = 31; // full detail for logfile
            DWORD dwTimeInt = static_cast <DWORD> (NMTime1->TimeInt);
            qwTimeInt = dwTimeInt * 10000000ui64;
            qwFinal = 94354848000000000ui64 + qwTimeInt + qBumpUp;
            uli.QuadPart = qwFinal;
            CopyMemory(&ft, &uli, sizeof(FILETIME));
            FileTimeToSystemTime(&ft, &st);
            if(SetSystemTime(&st)){
                LogWrite("Systemtime adjustment OK");
                if((GetIniString("terminate", "0") == "1"))
                    Application->Terminate();
                else
                    iDelay = StrToIntDef(GetIniString("sleep after success",
                            "1800"), 1800);
            }
            else{
                LogWrite("Systemtime adjustment failed");
                iDelay = StrToIntDef(GetIniString("sleep after failure",
                        "30"), 30);
            }
        }
        catch(...){
            LogWrite("NMTime exception");
        }
        Timer1->Enabled = true;
    }
}
//---------------------------------------------------------------------------
void TForm1::DoDialog(){
    if(Form2)
        return;
    Timer1->Enabled = false; // stall processing
    Form2 = new TForm2(Application);
    Form2->AddHoursUD->Position =
            static_cast <short> (StrToIntDef(GetIniString("extra hours",
            "0"), 0));
    Form2->AddMinsUD->Position =
            static_cast <short> (StrToIntDef(GetIniString("extra minutes",
            "0"), 0));
    Form2->AddSecsUD->Position =
            static_cast <short> (StrToIntDef(GetIniString("extra seconds",
            "0"), 0));
    Form2->DelayAfterFailureUD->Position =
            static_cast <short> (StrToIntDef(GetIniString("sleep after failure",
            "120"), 120));
    Form2->DelayAfterSuccessUD->Position =
            static_cast <short> (StrToIntDef(GetIniString("sleep after success",
            "3600"), 3600));
    Form2->InitialDelayUD->Position =
            static_cast <short> (StrToIntDef(GetIniString("initial delay",
            "0"), 0));
    Form2->LocalIPEdit->Text = GetIniString("local IP", "127.0.0.1");
    Form2->TimeoutUD->Position =
            static_cast <short> (StrToIntDef(GetIniString("NM timeout",
            "30"), 30));
    Form2->TimerUD->Position =
            static_cast <short> (StrToIntDef(GetIniString("timer interval",
            "5"), 5));
    Form2->AutostartCheckBox->Checked = (GetIniString("autostart", "1") == "1");
    Form2->OnceCheckBox->Checked = (GetIniString("terminate", "0") == "1");
    Form2->HostComboBox->Items->CommaText = GetIniString("daytime hosts",
            "time-a.nist.gov");
    Form2->HostComboBox->ItemIndex = StrToIntDef(GetIniString("current host",
            "0"), 0);
    if(Form2->ShowModal() == mrOk){
        PutIniString("current host", IntToStr(Form2->HostComboBox->ItemIndex));
        PutIniString("extra hours", Form2->AddHoursEdit->Text);
        PutIniString("extra minutes", Form2->AddMinsEdit->Text);
        PutIniString("extra seconds", Form2->AddSecsEdit->Text);
        PutIniString("sleep after failure", Form2->DelayAfterFailureEdit->Text);
        PutIniString("sleep after success", Form2->DelayAfterSuccessEdit->Text);
        PutIniString("initial delay", Form2->InitialDelayEdit->Text);
        PutIniString("local IP", Form2->LocalIPEdit->Text);
        PutIniString("NM timeout", Form2->TimeoutEdit->Text);
        PutIniString("timer interval", Form2->TimerEdit->Text);
        PutIniString("autostart", (Form2->AutostartCheckBox->Checked ? "1":"0"));
        PutIniString("terminate", (Form2->OnceCheckBox->Checked ? "1":"0"));
        Timer1->Interval = StrToIntDef(GetIniString("timer interval",
                "5"), 5) * 1000;
        Install(Form2->AutostartCheckBox->Checked);
        ComputeBumpUp();
        iDelay = StrToIntDef(Form2->InitialDelayEdit->Text, 0);
        localIP = Form2->LocalIPEdit->Text;
    }
    delete Form2;
    Form2 = 0;
    Timer1->Enabled = true; // continue processing
}
//---------------------------------------------------------------------------
void TForm1::ComputeBumpUp(){
    DWORD dwHours = StrToIntDef(GetIniString("extra hours", "0"), 0);
    DWORD dwMins = StrToIntDef(GetIniString("extra minutes", "0"), 0);
    DWORD dwSecs = StrToIntDef(GetIniString("extra seconds", "0"), 0);
    qBumpUp = (dwSecs + (dwMins*60ui64) + (dwHours*3600ui64)) * 10000000ui64;
}
//---------------------------------------------------------------------------

